/*
** ntk_hypothesis_testing.cc
** Login : <nicolas.burrus@ensta.fr>
** Started on  Thu Jul 27 15:03:00 2006 Nicolas Burrus
** $Id$
** 
** Copyright (C) 2006, 2007 Nicolas Burrus
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
** 
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
** 
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/

#include "hypothesis_testing.h"

#include <ntk/utils/math.h>
#include <ntk/utils/debug.h>

#include <vcl_numeric.h>
#include <vcl_cmath.h>
#include <vcl_algorithm.h>

namespace ntk
{

  double deviation_log_pfa(double dev, double n, double gdev, double gmu4)
  {
    if (n < 4) return 0.;
    if (dev >= gdev) return 0.;

  // Ensures that the estimation is valid.
    if ((gmu4/(gdev*gdev*gdev*gdev) < 1.3)) return 0;

    double vs = ((n-1)/(n*n*n))*((n-1)*gmu4-((n-3)*gdev*gdev*gdev*gdev));
    vs = sqrt(vs);
    double a = dev*dev - ((n-1)/n)*gdev*gdev;
    a /= vs;
  
    double pfa = lnnormcdf(a, 0, 1);
    ntk_assert(pfa < FLT_MAX, "Invalid float values.");
    return pfa;
  }
  
  vcl_vector< vcl_vector< vcl_vector<double> > > generate_mann_whitney_values()
  {
    vcl_vector< vcl_vector< vcl_vector<double> > > result (10);

    for (size_t n2 = 0; n2 < result.size(); ++n2)
    {
      result[n2].resize(result.size());
      for (size_t n1 = 0; n1 < result.size(); ++n1)
      {
        result[n2][n1].resize(result.size()*result.size());
        for (size_t u = 0; u < result.size()*result.size(); ++u)
        {
          if (n1 == 0 || n2 == 0)
          {
            result[n2][n1][u] = 1.;
            continue;
          }
        
          double w1 = ((double)n1)/(n1+n2);
          double t1 = w1 * result[n2][n1-1][u];
          double t2 = (u<n1) ? 0. : (1.-w1) * result[n2-1][n1][u-n1]; 
          result[n2][n1][u] = t1 + t2;          
        }
      }
    }
  
    return result;
  }

  template <class T>
  double compute_mann_whitney_log_pfa(const vcl_vector<T>& d1,
                                      const vcl_vector<T>& d2,
                                      double n1, 
                                      double n2)
  {
    static vcl_vector< vcl_vector< vcl_vector<double> > > mw_values 
      = generate_mann_whitney_values();
  
    assert(d1.size() == d2.size());
    if (n1 < 0) n1 = vcl_accumulate(d1.begin(), d1.end(), 0.);
    if (n2 < 0) n2 = vcl_accumulate(d2.begin(), d2.end(), 0.);
    double cur_rank = 1;
    double r1 = 0.;
    uint32 size = d1.size();
    for (uint32 i = 0; i < size; ++i)
    {
      const double d1_i = d1[i];
      const double d2_i = d2[i];
      double delta_rank = d1_i + d2_i;
      r1 += 0.5 * ((delta_rank-1) + 2*cur_rank) * d1_i;
      cur_rank += delta_rank;
    }
    double u = n1*n2 + 0.5*n1*(n1+1) - r1;
    u = vcl_min(u, n1*n2-u);
  
    double r;
  
    if (vcl_max(n1, n2) < mw_values.size())
      r = log(mw_values[(int)n2][(int)n1][(int)u]);
    else
      r = lnnormcdf(u, 0.5*n1*n2, sqrt((n1*n2*(n1+n2+1))/12.));
    return r;
  }

  template double compute_mann_whitney_log_pfa<double>(const vcl_vector<double>& d1, 
                                                       const vcl_vector<double>& d2,
                                                       double n1, double n2);

  template double compute_mann_whitney_log_pfa<int>(const vcl_vector<int>& d1, 
                                                    const vcl_vector<int>& d2,
                                                    double n1, double n2);

  template double compute_mann_whitney_log_pfa<unsigned>(const vcl_vector<unsigned>& d1, 
                                                         const vcl_vector<unsigned>& d2,
                                                         double n1, double n2);
  
} // end of ntk
