/*
** moments.cc
** Login : <nicolas.burrus@ensta.fr>
** Started on  Sun Oct 15 18:32:24 2006 Nicolas Burrus
** $Id$
**
** Copyright (C) 2006, 2007 Nicolas Burrus
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/

#include "moments.h"
#include <ntk/utils/iostream.h>
#include <ntk/utils/debug.h>

#include <vcl_numeric.h>
#include <vcl_limits.h>
#include <vcl_algorithm.h>
#include <vcl_iterator.h>

namespace ntk
{

  vcl_vector<double>
  fill_linear_values_from_control_points(const vcl_vector<Point>& control_points)
  {
    vcl_copy(stl_bounds(control_points), vcl_ostream_iterator<Point>(vcl_clog, " "));
    vcl_vector<double> distrib(unsigned(control_points.back().getX()+1));
    int cp_i = 0;
    foreach_idx(i, distrib)
    {
      while (control_points[cp_i+1].getX() < i)
      {
        ntk_assert((int)cp_i < (int)control_points.size()-1, "Out of range.");
        cp_i++;
      }
      Point left_p = control_points[cp_i];
      Point right_p = control_points[cp_i+1];
      double l = (i-left_p.getX())/(right_p.getX()-left_p.getX());
      ntk_assert(l <= 1.0, "l is too big.");
      double r = 1.-l;
      double v = left_p.getY()*r + right_p.getY()*l;
      distrib[i] = v;
    }
    return distrib;
  }

  void keep_local_maxima(vcl_vector<Point>* maxima, const vcl_vector<double>& histogram)
  {
    ntk_assert(maxima != 0, "Maxima is null.");
    ntk_assert(histogram.size() >= 2, "Histogram to small.");
    double prev_value = histogram[0];
    maxima->push_back(Point(0, histogram[0]));
    for (size_t i = 1; i < histogram.size()-1; ++i)
    {
      double next_value = i < histogram.size()-1 ? histogram[i+1] : -FLT_MAX;
      if (histogram[i] >= prev_value && histogram[i] >= next_value)
        maxima->push_back(Point(i, histogram[i]));
      prev_value = histogram[i];
    }
    maxima->push_back(Point(histogram.size()-1, histogram.back()));
  }

  vcl_vector<Point> make_unimodal_distribution(const vcl_vector<Point>& points)
  {
    double max_value = -FLT_MAX;
    foreach_idx(i, points)
      if (points[i].getY() > max_value)
        max_value = points[i].getY();

    vcl_vector<Point> control_points;
    double current_max = -FLT_MAX;
    int i = points.size()-1;
    for (i = points.size()-1; i > 0 && points[i].getY() < max_value; --i)
    {
      if (points[i].getY() > current_max)
      {
        current_max = points[i].getY();
        control_points.insert(control_points.begin(), points[i]);
      }
    }
    int last_index = i;
    control_points.insert(control_points.begin(), points[i]);

    vcl_vector<Point> control_points_fwd;
    current_max = -FLT_MAX;
    for (i = 0; i < last_index; ++i)
    {
      if (points[i].getY() > current_max)
      {
        current_max = points[i].getY();
        control_points_fwd.push_back(points[i]);
      }
    }
    control_points_fwd.insert(control_points_fwd.end(), stl_bounds(control_points));
    return control_points_fwd;
  }

  vcl_vector<double> distrib_to_bimodal_linear(const vcl_vector<double>& factors)
  {
    double max_value = -FLT_MAX;
    foreach_idx(i, factors) if (factors[i] > max_value) max_value = factors[i];

    vcl_vector<Point> control_points;
    double current_max = -FLT_MAX;
    int i = factors.size()-1;
    for (i = factors.size()-1; i > 0 && factors[i] < max_value; --i)
    {
      if (factors[i] > current_max)
      {
        current_max = factors[i];
        control_points.insert(control_points.begin(), Point(i, factors[i]));
      }
    }
    int last_index = i;
    control_points.insert(control_points.begin(), Point(i, factors[i]));

    vcl_vector<Point> control_points_fwd;
    current_max = -FLT_MAX;
    for (i = 0; i < last_index; ++i)
    {
      if (factors[i] > current_max)
      {
        current_max = factors[i];
        control_points_fwd.push_back(Point(i, factors[i]));
      }
    }
    control_points_fwd.insert(control_points_fwd.end(), stl_bounds(control_points));
    return fill_linear_values_from_control_points(control_points_fwd);
  }

} // end of ntk

