/*
** vil_binary_image.h
** Login : <nicolas.burrus@ensta.fr>
** Started on  Thu Jun  9 18:36:23 2005 Nicolas Burrus
** $Id$
** 
** Copyright (C) 2005, 2006 Nicolas Burrus
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
** 
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
** 
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/

#ifndef   	NTK_VXL_VIL_BINARY_IMAGE_H_
# define   	NTK_VXL_VIL_BINARY_IMAGE_H_

# include <ntk/core.h>

# include <vil/vil_border.h>
# include <vil/vil_image_view.h>
# include <vil/vil_load.h>
# include <vil/vil_save.h>
# include <vil/vil_smart_ptr.txx>
# include <vil/algo/vil_structuring_element.h>

# include <boost/dynamic_bitset.hpp>

namespace ntk {

  class vil_dynamic_bitset_ptr
  {
    //: Data
    boost::dynamic_bitset<>* data_;

    //: Reference count
    int ref_count_;

    //: Size
    size_t size_;

  public:
  
    vil_dynamic_bitset_ptr()
      : data_(0), ref_count_(0), size_(0)
    {}

    vil_dynamic_bitset_ptr(unsigned n, unsigned nb_elms)
      : data_(0), ref_count_(0), size_(0)
    {
      set_size(n, nb_elms);
    }

    //: Copy ctor
    vil_dynamic_bitset_ptr(const vil_dynamic_bitset_ptr& d)
      : data_(0), ref_count_(0), size_(0)
    {
      set_size(d.size(), 0);
      for (unsigned i = 0; i < size_; ++i)
        data_[i] = d.data_[i];
    }

    //: Assignment operator
    vil_dynamic_bitset_ptr& operator=(const vil_dynamic_bitset_ptr& d)
    {
      if (this==&d) return *this;
      set_size(d.size(), 0);
      for (unsigned i = 0; i < size_; ++i)
        data_[i] = d.data_[i];
      return *this;
    }    
    
    //: Increment reference count
    void ref() { ref_count_++; }
  
    //: Decrement reference count
    void unref()
    {
      assert (ref_count_ >0);
      ref_count_--;
      if (ref_count_==0)
      {
        delete [] data_; data_ = 0;
        delete this;
      }
    }
  
    //: Number of objects referring to this data
    int ref_count() const { return ref_count_; }

    //: Number of bitsets.
    int size() const { return size_; }

    //: Create empty space for n elements.
    //  Leave existing data untouched if the size is already n.
    void set_size(size_t n, unsigned nb_elms)
    {
      if (size_==n) return;
      delete [] data_;
      data_ = 0;
      if (n>0)
        data_ = new boost::dynamic_bitset<>[n];
      for (unsigned i = 0; i < n; ++i)
        data_[i].resize(nb_elms);
      size_ = n;
    }

    //: Pointer to data.
    boost::dynamic_bitset<>* data() { return data_;}
  
    //: Pointer to first element of data
    boost::dynamic_bitset<>* const_data() const { return data_;}
  };

} // end of ntk

typedef vil_smart_ptr<ntk::vil_dynamic_bitset_ptr> vil_dynamic_bitset_sptr;

namespace ntk {

  class vil_binary_image : public vil_image_view_base
  {
  public:
    typedef boost::dynamic_bitset<> bitset_type;

  private:
    vil_dynamic_bitset_sptr bitsets_sptr_;
    bitset_type* bitsets_;
    vcl_ptrdiff_t istep_;
    vcl_ptrdiff_t jstep_;
    mutable bool last_value_; // Cache value to avoid returning temporaries.

    //: Disconnect this view from the underlying data,
    void release_memory() { bitsets_sptr_ = 0; }

  public:
    vil_binary_image()
      : bitsets_sptr_(0), bitsets_(),
        istep_(0), jstep_(0)
    {}

    vil_binary_image(unsigned width, unsigned height)
      : bitsets_sptr_(0), bitsets_(),
        istep_(0), jstep_(0)
    {
      set_size(width, height, 1);
    }

    vil_binary_image(unsigned width, unsigned height, unsigned n_planes)
      : bitsets_sptr_(0), bitsets_(0),
        istep_(0), jstep_(0)
    {
      set_size(width, height, n_planes);
    }

    vil_binary_image(const vil_dynamic_bitset_sptr& sptr, 
                     const bitset_type* bitsets,
                     unsigned width, unsigned height, unsigned n_planes)
      : vil_image_view_base (width, height, n_planes),
        bitsets_sptr_(const_cast<vil_dynamic_bitset_sptr&>(sptr)), 
        bitsets_(const_cast<bitset_type*>(bitsets)),
        istep_(1), jstep_(width)
    {}

    vil_binary_image(const vil_binary_image& rhs)
      : vil_image_view_base (rhs.ni(), rhs.nj(), rhs.nplanes()),
        bitsets_sptr_(0), bitsets_(),
        istep_(0), jstep_(0)
    {
      operator=( static_cast<vil_binary_image const&>(rhs) );
    }

    template <class T>
    vil_binary_image(const vil_image_view<T>& src)
    {
      deep_copy(src);
    }

    vil_binary_image& operator= (const vil_binary_image& rhs);

    //  Destructor
    virtual ~vil_binary_image() {}

    virtual vil_binary_image& clone() const 
    {
      vil_binary_image* tmp = new vil_binary_image();
      tmp->deep_copy(*this);
      return *tmp;
    }
    
    void load (const char* filename);

    void deep_copy(const vil_binary_image& src);

    template <class T>
    void deep_copy(const vil_image_view<T>& src);

    //: Export to a vil_image_view<T>
    template <class T>
    void copy_to(vil_image_view<T>& dest) const;

    //: The pixel type of this image
    typedef bool pixel_type;

    //: set_size current planes to width x height.
    // If already correct size, this function returns quickly
    virtual void set_size(unsigned width, unsigned height)
    { set_size(width, height, 1); }
  
    //: resize to width x height x n_planes.
    // If already correct size, this function returns quickly
    virtual void set_size(unsigned width, unsigned height, unsigned n_planes);

    boost::dynamic_bitset<>* bitsets () { return bitsets_; }
    const boost::dynamic_bitset<>* bitsets () const { return bitsets_; }

    boost::dynamic_bitset<>& bitset (int i = 0) { return bitsets_[i]; }
    const boost::dynamic_bitset<>& bitset (int i = 0) const { return bitsets_[i]; }
    const vil_dynamic_bitset_sptr& bitsets_sptr () const { return bitsets_sptr_; }

    //: Add this to your pixel pointer to get next i pixel.
    //  Note that istep() may well be negative; see e.g. vil_flip_lr
    inline vcl_ptrdiff_t istep() const { return istep_; }
    //: Add this to your pixel pointer to get next j pixel.
    //  Note that jstep() may well be negative; see e.g. vil_flip_ud
    inline vcl_ptrdiff_t jstep() const { return jstep_; }

    // Returns true if there is no data.
    bool operator!() const { return bitsets_sptr_ == 0; }
    
    inline bool in_range (int i, int j) const
    { return i < int(ni_) && j < int(nj_) && i > -1 && j > -1; }

    inline bool in_range (int i, int j, int p) const
    { 
      return i < int(ni_) && j < int(nj_) && p < int(nplanes_)
             && i > -1 && j > -1 && p > -1;
    }

    unsigned count () const;

    void fill (bool value);

    inline const bool& operator()(unsigned i, unsigned j) const
    {
      int idx = j*jstep_+i*istep_;
      last_value_ = (*bitsets_)[idx];
      return last_value_;
    }

    inline const bool& operator()(unsigned i, unsigned j, unsigned p) const
    {
      int idx = j*jstep_ + i*istep_;
      last_value_ = bitsets_[p][idx];
      return last_value_;
    }

    inline const bool& operator[](unsigned idx) const
    {
      last_value_ = (*bitsets_)[idx];
      return last_value_;
    }

    inline void set(unsigned idx, bool value)
    {
      (*bitsets_)[idx] = value;
    }

    inline void set(unsigned i, unsigned j, bool value)
    {
      int idx = j*jstep_ + i*istep_;
      (*bitsets_)[idx] = value;
    }

    inline void set(unsigned i, unsigned j, unsigned p, bool value)
    {
      int idx = j*jstep_ + i*istep_;
      bitsets_[p][idx] = value;
    }

    void flip()
    {
      for (unsigned p = 0; p < nplanes_; ++p)
        bitsets_[p].flip();
    }

    vil_binary_image& operator|= (const vil_binary_image& rhs);
    vil_binary_image& operator&= (const vil_binary_image& rhs);
    vil_binary_image& operator^= (const vil_binary_image& rhs);
    bool operator== (const vil_binary_image& rhs) const;

    virtual void print (vcl_ostream&) const {}
    virtual vcl_string is_a () const { return "vil_binary_image"; }
    virtual vil_pixel_format pixel_format() const
    { return vil_pixel_format_of(bool()); }
  };

  vil_binary_image 
  operator| (const vil_binary_image& lhs, const vil_binary_image& rhs);
  
  vil_binary_image 
  operator& (const vil_binary_image& lhs, const vil_binary_image& rhs);
  
  vil_binary_image 
  operator^ (const vil_binary_image& lhs, const vil_binary_image& rhs);
  
  bool vil_save(const vil_binary_image &im, char const* filename);

  vil_binary_image vil_plane(const vil_binary_image& im, unsigned nplane);

  void vil_binary_erode(const vil_binary_image& src_image, 
                        vil_binary_image& dest_image,
                        const vil_structuring_element& element);
  
  void vil_binary_erode(const vil_binary_image& src_image, 
                        vil_binary_image& dest_image,
                        const vil_structuring_element& element,
                        vil_border<vil_binary_image> border);

  void vil_binary_dilate(const vil_binary_image& src_image, 
                         vil_binary_image& dest_image,
                         const vil_structuring_element& element);
  
  void vil_binary_dilate(const vil_binary_image& src_image, 
                         vil_binary_image& dest_image,
                         const vil_structuring_element& element,
                         vil_border<vil_binary_image> border);

  /*-------------------------.
  | Template implementations |
  `-------------------------*/

  //: Export to a vil_image_view<bool>.
  template <class T>
  void
  vil_binary_image::copy_to(vil_image_view<T>& dest) const
  {
    dest.set_size(ni(), nj(), nplanes());
    if (dest.is_contiguous())
    {
      unsigned plane_size = ni()*nj();
      T* p_dest = dest.top_left_ptr();
      for (unsigned p = 0; p < nplanes_; ++p, ++p_dest)
      {
        for (unsigned i = 0; i < plane_size; ++i, ++p_dest)
          *p_dest = (T) bitsets_[p][i];
      }
      return;
    }

    const vcl_ptrdiff_t s_planestep = dest.planestep();
    const vcl_ptrdiff_t s_istep = dest.istep();
    const vcl_ptrdiff_t s_jstep = dest.jstep();

    // Do a deep copy
    // This is potentially inefficient
    T* dest_data = dest.top_left_ptr();
    for (unsigned int p=0;p<nplanes_;++p,dest_data += s_planestep)
    {
      unsigned i = 0;
      T* dest_row = dest_data;
      for (unsigned int j=0;j<nj_;++j,dest_row += s_jstep, ++i)
      {
        T* sp = dest_row;
        for (unsigned int i=0;i<ni_;++i,sp+=s_istep, ++i) 
          *sp = bitsets_[p][i];
      }
    }  
  }

  template <class T>
  void
  vil_binary_image::deep_copy(const vil_image_view<T>& src)
  {
    set_size(src.ni(), src.nj(), src.nplanes());
    if (src.is_contiguous())
    {
      int plane_size = ni()*nj();
      const T* p_src = src.top_left_ptr();
      bitset_type* bset;
      for (unsigned p = 0; p < nplanes_; ++p, ++p_src)
      {
        bset = bitsets_ + p;
        for (int i = 0; i < plane_size; ++p_src, ++i)
          bset->set(i, (*p_src != 0));
      }
      return;
    }

    const vcl_ptrdiff_t s_planestep = src.planestep();
    const vcl_ptrdiff_t s_istep = src.istep();
    const vcl_ptrdiff_t s_jstep = src.jstep();

    // Do a deep copy
    // This is potentially inefficient
    const T* src_data = src.top_left_ptr();
    for (unsigned int p=0;p<nplanes_;++p,src_data += s_planestep)
    {
      unsigned i = 0;
      const T* src_row = src_data;
      for (unsigned int j=0;j<nj_;++j,src_row += s_jstep, ++i)
      {
        const T* sp = src_row;
        for (unsigned int i=0;i<ni_;++i,sp+=s_istep, ++i)
          bitsets_[p].set(i, (*sp != 0));
      }
    }  
  }
  
} // end of ntk

#endif	    /* !NTK_VXL_VIL_BINARY_IMAGE_H_ */
