//
// ac_explorer.cc
//
// Author: Nicolas Burrus <nicolas.burrus@ensta.fr>, (C) 2007
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
//

#include "ac_explorer.h"
#include "ac_segmentor.h"

#include <boost/multi_index_container.hpp>
#include <boost/multi_index/ordered_index.hpp>
#include <boost/multi_index/hashed_index.hpp>
#include <boost/multi_index/identity.hpp>
#include <boost/multi_index/mem_fun.hpp>
#include <boost/multi_index/member.hpp>
#include <boost/multi_index/composite_key.hpp>

#include <ntk/ntk_with_macros.h>
#include <vcl_set.h>

using namespace boost;
using namespace boost::multi_index;

String MergeBestExplorer ::
name() const
{
  return T("MergeBestExplorerInit") + m_initializer->name() + T("V1");
}

void MergeBestExplorer ::
initialize(AcSelector& selector)
{
  AcExplorer::initialize(selector);
  m_edges.clear();

  m_regions_to_merge.clear();
  m_regions_correspondances.clear();
  
  m_initializer->segment(segmentor().getImageToSegment());
  segmentor().importRegionsFrom(*m_initializer);
  m_initializer->clear();

  selector.segmentorInitialized();
  buildEdges();
}

#if 1
void MergeBestExplorer ::
run()
{
  int nb_it = 0;
  while (true)
  {
    ntk_dbg_print((int)m_edges.size(), 3);

    if (m_edges.size() == 0) break;

    edge_set_by_score_type::iterator lower_it = getByScore(m_edges).begin();

    if (m_max_regions < 0)
    {
      if (lower_it->isMeaningful()) break;
    }
    else
    {
      if ((int)segmentor().nbRegions() < m_max_regions) break;
    }

    ntk_dbg(3) << "[" << lower_it->r1() << "] <== [" << lower_it->r2() << "] ("
        << lower_it->score() << ")";

    mergeRegions(*lower_it);
    if (segmentor().refreshRate() > 0 && (nb_it % segmentor().refreshRate() == 0))
    {
      segmentor().updateCurrentSegmentation();
    }
    ++nb_it;
  }

  selector().finalize();
}
#else
                      
void MergeBestExplorer :: 
run()
{
  int nb_it = 0;
  while (true)
  {
    ntk_dbg_print((int)m_edges.size(), 3);
    if (m_edges.size() == 0) break;

    edge_set_by_score_type::iterator lower_it = getByScore(m_edges).begin();
    if (m_max_regions < 0)
    {
      if (lower_it->isMeaningful()) break;
    }
    else
    {
      if ((int)segmentor().nbRegions() < m_max_regions) break;
    }
    
    if (m_regions_to_merge.size() == 0)
    {
      ntk_dbg(0) << "Refilling m_regions_to_merge.";
      // Refill it.
      edge_set_by_score_type::iterator current_it = lower_it;
      while (current_it != getByScore(m_edges).end())
      {
        m_regions_to_merge.push_back(*current_it);
        ++current_it;
      }
      ntk_dbg(0) << "m_regions_to_merge refilled.";
    }
    
    RegionEdge edge = m_regions_to_merge.front(); m_regions_to_merge.pop_front();
    uint32 r1 = getRealRegionId(edge.r1());
    uint32 r2 = getRealRegionId(edge.r2());
    if (r1 == r2) continue;
    if (r1 > r2) vcl_swap(r1, r2);
    edge_set_by_label_type::iterator edge_it =
        getByLabel(m_edges).find(boost::make_tuple(r1, r2));
    ntk_assert(edge_it != getByLabel(m_edges).end(), "Bad regions");
    if (edge_it->isMeaningful()) continue;
    
    mergeRegions(*edge_it);
    if (segmentor().refreshRate() > 0 && (nb_it % segmentor().refreshRate() == 0))
    {
      segmentor().updateCurrentSegmentation();
    }
    ++nb_it;
  }
  selector().finalize();
}
                      
#endif
                      
void MergeBestExplorer ::
mergeRegions(const RegionEdgeWithScore& edge)
{
  Region& r1 = segmentor().getRegionRef(edge.r1());
  Region& r2 = segmentor().getRegionRef(edge.r2());

  // Remove links to r2.
  removeEdgesTo(r2);

  // Merge the two regions.
  segmentor().mergeR2IntoR1(r1.id(), r2.id());

  // Tell the selector something has changed.
  selector().r2MergedIntoR1(r1.id(), r2.id());

  // Update links with neighbors of the new big region.
  for (Region::neighbors_const_iterator it = r1.neighborsBegin();
       it != r1.neighborsEnd(); ++it)
  {
    ntk_dbg(5) << "neighbor of [" << r1.id() << "]: " << *it;
    addEdge(r1.id(), *it);
  }

  m_regions_correspondances[r2.id()] = r1.id();  
  segmentor().removeRegion(r2); 
}

void MergeBestExplorer ::
addEdge(uint32 r1, uint32 r2)
{
  double score; bool is_meaningful;
  selector().getDifference(r1,r2,&score, &is_meaningful);
  RegionEdgeWithScore e (r1, r2, score, is_meaningful);
  vcl_pair<edge_set_by_label_type::iterator,bool> insert_result
      = getByLabel(m_edges).insert(e);

  if (insert_result.second == false)
  {
    // an element already exists with the same key.
    // if it was meaningful, force it to remain meaningful.
    if (m_keep_all_meaningful_links && insert_result.first->isMeaningful())
    {
      if (!e.isMeaningful())
        ntk_dbg_print("Enforcing a meaningful link", 2);
      e.setMeaningful(true);
    }
    getByLabel(m_edges).replace(insert_result.first, e);
  }
}

void MergeBestExplorer ::
removeEdgesTo(const Region& r)
{
  for (Region::neighbors_const_iterator it = r.neighborsBegin();
       it != r.neighborsEnd(); ++it)
  {
    segmentor().getRegionRef(*it).removeNeighbor(r.id());
    RegionEdgeWithScore tmp (*it, r.id()); // make sure the edge is well built.
    edge_set_by_label_type::iterator edge_it =
        getByLabel(m_edges).find(boost::make_tuple(tmp.r1(), tmp.r2()));
    ntk_assert(edge_it != getByLabel(m_edges).end(), "Inexistant edge.");
    getByLabel(m_edges).erase(edge_it);
  }
}

void MergeBestExplorer ::
buildEdges()
{
  // FIXME: use region links !
  int max_id = 0;
  vil_image_view<int32> label_image = segmentor().getLabelImage();
  for_all_xy(label_image)
  {
#define add_edge_if_different(im, otherx, othery) \
    if (im.in_range(otherx,othery) && im(otherx,othery) != current_id) \
    addEdge(current_id, im(otherx,othery));

    int32 current_id = label_image(x,y);
    max_id = vcl_max(current_id, max_id);
    add_edge_if_different(label_image, x+1, y);
    add_edge_if_different(label_image, x,   y+1);
    add_edge_if_different(label_image, x+1, y+1);
    add_edge_if_different(label_image, x+1, y-1);
  }
  m_regions_correspondances.resize(max_id+1);
  foreach_idx(i, m_regions_correspondances)
    m_regions_correspondances[i] = i;
}

uint32 MergeBestExplorer ::
getRealRegionId(uint32 r)
{
  while (m_regions_correspondances[r] != r)
    r = m_regions_correspondances[r];
  return r;
}
