//
// ac_learn_process.cc
// 
// Author: Nicolas Burrus <nicolas.burrus@ensta.fr>, (C) 2007
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
//

#include "ac_learn_process.h"
#include "ac_segmentor.h"
#include "ac_selector.h"
#include "ac_threshold_database.h"

#include <ntk/ntk_with_macros.h>
#include <vil/algo/vil_gauss_filter.h>

using namespace ntk;

static void interpolate_missing_values(vcl_vector<double>& values,
                                       vcl_vector<bool>& mask)
{
  bool has_missing_value = true;
  while (has_missing_value)
  {
    has_missing_value = false;
    for (uint32 k = 0; k < values.size(); ++k)
    {
      if (mask[k]) continue;
      bool next_is_filled = k < values.size()-1 && mask[k+1];
      bool prev_is_filled = k >= 1 && mask[k-1];            
      if (next_is_filled && prev_is_filled) 
      {
        mask[k] = true;
        values[k] = (values[k+1]+values[k-1])/2.0;
      }
      else if (next_is_filled)
      {
        mask[k] = true;
        values[k] = values[k+1];
      }
      else if (prev_is_filled)
      {
        mask[k] = true;
        values[k] = values[k-1];
      }
      else
      {
        has_missing_value = true;
      }
    }
  }
}

/***********************************************************************/
/***********************************************************************/

AcLearnerPtr AcLearnProcess ::
createLearner(const PropertySet& properties)
{
  if (!properties.containsKey(T("learning-learner")))
    ntk_throw_exception("No learner attribute !");
  
  String learner = properties.getValue(T("learning-learner"));
  
  if (learner == T("max")) 
    return AcLearnerPtr(new AcMaxLearner(nbCategories()));
  
  if (learner == T("mean")) 
    return AcLearnerPtr(new AcMeanLearner(nbCategories()));
  
  ntk_throw_exception("Invalid learner.");
  return AcLearnerPtr();
}

void AcLearnProcess::
learn()
{
  m_learner->initialize();
  if (m_listener) m_segmentor->addChangeListener(m_listener);
  m_segmentor->startLearningMode(*m_learner);
  vnl_random rgen;
  for (uint32 iteration = 0; iteration < m_nb_iterations; ++iteration)
  {
    ntk_log() << (String)T("== [IMAGE ") << iteration << (String)T("] ==");

    vcl_vector<double> distrib(256, 0);
    // foreach_idx(i, distrib) distrib[i] = normpdf(i, 127, 10);
    foreach_idx(i, distrib) distrib[i] = 1.0;
    normalize_distrib(distrib);
    
    m_learner->newImage();
    {
      const ScopedLock lock(m_critical_section); 
      // m_source_image = well_sampled_noise(m_width, m_height, distrib, rgen);
      m_source_image.set_size(m_width, m_height);
      for_all_xy(m_source_image) m_source_image(x,y) = draw_from_distrib(distrib, rgen);
    }

    if (m_listener) m_listener->changeListenerCallback(this);
    if (shouldStop()) return;

    m_segmentor->segment(m_source_image);

    {
      const ScopedLock lock(m_critical_section);
      m_current_thresholds = m_learner->getThresholds();
    }
  }
  m_segmentor->stopLearningMode();
  
  m_learner->computeThresholds();
  
  {
    const ScopedLock lock(m_critical_section);
    m_current_thresholds = m_learner->getThresholds();
  }
  if (m_listener)
  {
    m_listener->changeListenerCallback(this);  
  }

  saveResultsToDatabase();
}

void AcLearnProcess ::
saveResultsToDatabase()
{
  const vcl_vector<double>& thresholds = m_learner->getThresholds();
  
  XmlElement root (T("root"));
  XmlElement* e = new XmlElement(T("thresholds"));
  e->setAttribute(T("name"), m_segmentor->name());
  e->setAttribute(T("loader"), "SizeDependent");
  e->setAttribute(T("width"), (int)m_width);
  e->setAttribute(T("height"), (int)m_height);
  
  String values;
  values << thresholds[0];
  for (uint32 i = 1; i < thresholds.size(); ++i)
    values << " " << thresholds[i];
  
  XmlElement* values_element = new XmlElement("values");
  values_element->setAttribute(T("width"), (int)m_width);
  values_element->setAttribute(T("height"), (int)m_height);
  values_element->addChildElement(XmlElement::createTextElement(values));
  e->addChildElement(values_element);
  root.addChildElement(e);
  m_segmentor->selector()->thresholdDatabase()->loadFromXmlElement(root);
}

/***********************************************************************/
/***********************************************************************/

void AcMaxLearner :: 
initializeImpl()
{
  m_raw_values.resize(nbCategories());
}

void AcMaxLearner :: 
addDistance(uint32 couple_category, double value)
{
  m_raw_values[couple_category].push_back(value);
  m_thresholds[couple_category] = vcl_max(value, m_thresholds[couple_category]);
}

void AcMaxLearner ::
computeThresholds()
{
  vcl_vector<bool> mask(m_thresholds.size(), true);
  
  for (uint32 k = 0; k < m_raw_values.size(); ++k)
  {
    if (m_raw_values[k].size() == 0)
    {
      // filled later by interpolation.
      mask[k] = false;
      continue;
    }
    
    m_thresholds[k] = m_raw_values[k][0];
    for (uint32 i = 1; i < m_raw_values[k].size(); ++i)
      m_thresholds[k] = vcl_max(m_thresholds[k], m_raw_values[k][i]);  
  }
  
  interpolate_missing_values(m_thresholds, mask);
  if (0) // disabled
  {
    vcl_ofstream f("learned_raw.plot");
    foreach_idx(k, m_thresholds) f << k << " " << m_thresholds[k] << "\n";
  }
  
  vcl_vector<Point> maxima;
  keep_local_maxima(&maxima, m_thresholds);
  m_thresholds = fill_linear_values_from_control_points(maxima);
  m_thresholds = distrib_to_bimodal_linear(m_thresholds);  
}

/***********************************************************************/
/***********************************************************************/

void AcMeanLearner::MaxImageValues::
addSorted(uint32 idx, double value)
{
  ntk_assert(idx < values.size(), "Invalid index.");
  if (values[idx].size() <= number_of_values_to_keep()
      || *values[idx].begin() < value)
  {
    values[idx].insert(value);
    if (values[idx].size() > number_of_values_to_keep())
      values[idx].erase(values[idx].begin());
  }
}
      
void AcMeanLearner :: 
initializeImpl()
{
  m_raw_values.clear();
}

void AcMeanLearner :: 
newImage()
{
  m_raw_values.push_back(MaxImageValues(nbCategories()));
}


void AcMeanLearner :: 
addDistance(uint32 couple_category, double value)
{
  ntk_assert(m_raw_values.size() > 0, "You should call newImage().");
  MaxImageValues& cur_values = m_raw_values.back();
  cur_values.addSorted(couple_category, value);
  m_thresholds[couple_category] = vcl_max(value, m_thresholds[couple_category]);
}

void AcMeanLearner :: 
optimizeThreshold(uint32 k)
{
  const uint32 max_iterations = 1000;
  const double global_confidence = 0.99;
  const double confidence_per_k = exp(log(0.99)/nbCategories());
  const double Q = m_raw_values.size();
  ntk_assert(Q >= 2, "Learning must be done with at least two images.");
  const double t_value = StudentDistrib(Q-1).invCdf()(confidence_per_k);
  
  double mu_upper_bound = FLT_MAX;
  for (uint32 iteration = 0; iteration < max_iterations; ++iteration)
  {  
    double mk = 0;
    double sk = 0;
    vcl_vector<double> nb_selected_in_each_image(m_raw_values.size(), 0);
    for (uint32 im = 0; im < m_raw_values.size(); ++im)
    {    
      const MaxImageValues& image_values = m_raw_values[im];
      if (image_values.values[k].size() == 0) continue;
      vcl_multiset<double>::const_reverse_iterator it;
      uint32 nb_selected = 0;
      for (it = image_values.values[k].rbegin(); 
           it != image_values.values[k].rend() && *it >= m_thresholds[k];
           ++it)
      {
        nb_selected += 1;
      }
      nb_selected_in_each_image[im] = nb_selected;
    }

    values_moments(stl_bounds(nb_selected_in_each_image), 
                   nb_selected_in_each_image.size(),
                   &mk, &sk);
    sk = vcl_max(sk, 1.0/nb_selected_in_each_image.size());
    mu_upper_bound = mk + (t_value*sk)/sqrt(Q-1);
    ntk_dbg(2) << "(mean,dev) of selected couples for " << k << ": (" 
        << mk << ", " << sk << ") => muk < " << mu_upper_bound;
    
    if (mu_upper_bound <= (m_epsilon/nbCategories())) break; // good results.
    if (flt_eq(mk, 0.0)) break; // Cannot improve results.

    m_thresholds[k] *= 1.01;
  }
  
  if (mu_upper_bound > (m_epsilon/nbCategories()))
  {
    ntk_log() << (String)T("*** Warning: upper bound for mu is greater than epsilon/K: ")
              << (String)T("k=") << k 
              << (String)T(" uppermu=") << mu_upper_bound;
  }
}

void AcMeanLearner ::
computeThresholds()
{
  m_thresholds.resize(nbCategories());
  vcl_fill(stl_bounds(m_thresholds), 0);
  ntk_dbg_print((int)m_thresholds.size(), 1);
  
  vcl_vector<bool> mask(m_thresholds.size(), true);
  
  vcl_vector<double> max_mean (m_thresholds.size(), 0);
  vcl_vector< vcl_multiset<double> > max_values (m_thresholds.size());
  for (uint32 im = 0; im < m_raw_values.size(); ++im)
  {    
    const MaxImageValues& image_values = m_raw_values[im];
    foreach_idx(k, image_values.values)
    {
      if (image_values.values[k].size() == 0) continue;
      max_values[k].insert(*image_values.values[k].rbegin());
    }
  }
  
  foreach_idx(k, max_values)
  {
    vcl_multiset<double>::const_reverse_iterator it;
    uint32 nb_values;
    for (it = max_values[k].rbegin(), nb_values = 0;
         it != max_values[k].rend();
         ++it, ++nb_values)
    {
      max_mean[k] += *it;
    }
    if (nb_values == 0) 
    { 
      mask[k] = false; 
      continue;
    }
    max_mean[k] /= nb_values;
  }

  // Initialization.
  m_thresholds = max_mean;
  foreach_idx(k, max_mean)
  {
    if (mask[k] == false) continue;
    optimizeThreshold(k);
  }
  
  interpolate_missing_values(m_thresholds, mask);
  vcl_vector<double> raw_values = m_thresholds;
  if (0) // disabled
  {
    vcl_ofstream f("learned_raw.plot");
    foreach_idx(k, m_thresholds) f << k << " " << m_thresholds[k] << "\n";
  }
   
  // Smooth curve by removing local minima.
  vcl_vector<Point> maxima;
  keep_local_maxima(&maxima, m_thresholds);
  maxima = make_unimodal_distribution(maxima);
  if (0) // disabled
  {
    vcl_ofstream f("learned_maxima.plot");
    foreach_idx(i, maxima)
      f << maxima[i].getX() << " " << maxima[i].getY() << "\n";
  }
    
  m_thresholds = fill_linear_values_from_control_points(maxima);
  
  vcl_vector<Point> new_control_points;
  vcl_vector<Point>::const_iterator maxima_it = maxima.begin();
  foreach_idx(k, m_thresholds)
  {
    if (mask[k] == false) continue;
    if (maxima_it != maxima.end() && (flt_eq(maxima_it->getX(), k)))
    {
      new_control_points.push_back(*maxima_it);
      ++maxima_it;
    }
    else if (raw_values[k] > m_thresholds[k])
    {
      new_control_points.push_back(Point(k, raw_values[k]));
    }
  }
  m_thresholds = fill_linear_values_from_control_points(new_control_points);
  if (0) // disabled
  {
    vcl_ofstream f("learned.plot");
    foreach_idx(k, m_thresholds) f << k << " " << m_thresholds[k] << "\n";
  }
}
