//
// segmentor.cc
//
// Author: Nicolas Burrus <nicolas.burrus@ensta.fr>, (C) 2007
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
//

#include "segmentor.h"

#include <ntk/ntk_with_macros.h>
#include <vil/algo/vil_region_finder.h>

using namespace ntk;

void Segmentor::
initialize()
{
  m_regions.clear();
  m_label_image.clear();
  initializeImpl();
}

void
Segmentor :: getMeanImage(vil_image_view<uint8>& output,
                          bool with_borders) const
{
  vil_byte_image im = getImageToSegment();
  output.set_size(im.ni(), im.nj());
  for (region_const_iterator it = regionsBegin(); it != regionsEnd(); ++it)
  {
    const Region& r = *(it);

    double mean = 0;
    for (uint32 i = 0; i < r.size(); ++i) mean += im(r[i].x, r[i].y);
    mean /= r.size();

    for (uint32 i = 0; i < r.size(); ++i)
    {
      output(r[i].x, r[i].y) = vnl_math_rnd(mean);
    }
  }

  vil_byte_image tmp(output.ni()*2, output.nj()*2);
  for_all_xy(tmp) tmp(x,y) = output(x/2, y/2);
  output.deep_copy(tmp);
  if (with_borders)
  {
    for_all_xy(tmp)
    {
      unsigned v = tmp(x,y);
      if (has_different_neighbor_forward_8conn(tmp, x, y))
        output(x,y) = 0;
    }
  }
}

void Segmentor :: importFromLabelImage(const vil_image_view<int32>& label_image)
{
  clear();

  ntk_assert(label_image, "Null input label image.");

  vil_region_finder<int32>
      region_finder (label_image, vil_region_finder_8_conn);
  unsigned next_id = 0;
  for_all_xy(label_image)
  {
    if (region_finder.boolean_region_image()(x,y))
      continue; // already processed.
    vcl_vector<unsigned> ri,rj;
    region_finder.same_int_region(x,y,ri,rj);

    Region& r = createRegion(next_id);
    ++next_id;
    for (uint32 k = 0; k < ri.size(); ++k)
      r.addPoint(vil_point(ri[k], rj[k]));
  }
  createRegionLinksAndLabelImage();
}

void Segmentor :: createRegionLinks()
{
  ntk_assert(m_label_image, "Label image must exist.");

  for_all_xy(m_label_image)
  {
#define add_link_if_different(im, otherx, othery) \
    if (im.in_range(otherx,othery) && im(otherx,othery) != current_id) \
    { \
      int32 other_id = im(otherx,othery); \
      getRegionRef(current_id).addNeighbor(other_id); \
      getRegionRef(other_id).addNeighbor(current_id); \
    }

    int32 current_id = m_label_image(x,y);
    add_link_if_different(m_label_image, x+1, y);
    add_link_if_different(m_label_image, x,   y+1);
    add_link_if_different(m_label_image, x+1, y+1);
    add_link_if_different(m_label_image, x+1, y-1);
#undef add_link_if_different
  }
}

void Segmentor :: createRegionLinksAndLabelImage()
{
  // Create label image.
  vil_byte_image im = getImageToSegment();
  m_label_image.set_size(im.ni(), im.nj());
  for (region_const_iterator it = regionsBegin(); it != regionsEnd(); ++it)
  {
    const Region& r = *(it);
    for (uint32 i = 0; i < r.size(); ++i)
    {
      m_label_image(r[i].x, r[i].y) = r.id();
    }
  }

  createRegionLinks();
}

void Segmentor :: mergeR2IntoR1(uint32 r1_id, uint32 r2_id)
{
  Region& r1 = getRegionRef(r1_id);
  Region& r2 = getRegionRef(r2_id);
  r1.mergeWith(r2);
  for (Region::neighbors_const_iterator it = r2.neighborsBegin();
       it != r2.neighborsEnd(); ++it)
  {
    Region& r = getRegionRef(*it);
    r.removeNeighbor(r2_id);
    r.addNeighbor(r1_id);
  }
}
