//
// segmentor.h
//
// Author: Nicolas Burrus <nicolas.burrus@ensta.fr>, (C) 2007
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
//

#ifndef ACSEGMENTOR_segmentor_H
# define ACSEGMENTOR_segmentor_H

# include <ntk/ntk.h>
# include <vcl_map.h>

# include "region.h"

class Segmentor;
typedef boost::shared_ptr<Segmentor> SegmentorPtr;

class Segmentor : public ChangeBroadcaster
{
  public:
    typedef vcl_map<uint32,Region> region_map_type;
    typedef ntk::key_iterator<region_map_type::const_iterator> region_id_const_iterator;
    typedef ntk::value_iterator<region_map_type::const_iterator> region_const_iterator;

  public:
    Segmentor() : m_refresh_rate(-1) {}
    virtual ~Segmentor() {}

  public:
    virtual String name() const = 0;
    virtual void clear() { m_regions.clear(); }    
    
  public:
    virtual void segment(const ntk::vil_byte_image& image_to_segment)
    { 
      setImageToSegment(image_to_segment); 
      
      initialize();
      updateCurrentSegmentation();
      segmentImpl();      
      updateCurrentSegmentation();
      finalize();
      ntk_log() << "Final number of regions: " << nbRegions();
    }

    int refreshRate() const { return m_refresh_rate; }
    void setRefreshRate(int value) { m_refresh_rate = value; }
    
    void setImageToSegment(const ntk::vil_byte_image& image) { m_image_to_segment = image; }
    const ntk::vil_byte_image& getImageToSegment() const { return m_image_to_segment; }

    const vil_image_view<int32>& getLabelImage() const { return m_label_image; }
    
    virtual void getMeanImage(vil_image_view<uint8>& output,
                              bool with_borders = false) const;

    vil_image_view<uint8> getCurrentSegmentation() const 
    { ScopedLock locker(m_lock); return m_current_segmentation; }
    
    void updateCurrentSegmentation() 
    { ScopedLock locker(m_lock); getMeanImage(m_current_segmentation, true); sendChangeMessage(this); }
    
    uint32 nbRegions() const { return m_regions.size(); }
    const Region& getRegion(uint32 id) const
    {
      region_map_type::const_iterator it = m_regions.find(id);
      ntk_assert(it != m_regions.end(), "Wrong id.");
      return it->second;
    }
    
    virtual region_const_iterator regionsBegin() const { return m_regions.begin(); }
    virtual region_const_iterator regionsEnd() const { return m_regions.end(); }

  public:
    void importRegionsFrom(const Segmentor& rhs)
    { m_regions = rhs.m_regions; m_label_image = rhs.m_label_image; }

    void addRegion(const Region& r)
    {
      ntk_assert(!ntk::has_key(m_regions, r.id()), "Region already exists.");
      m_regions[r.id()] = r;
    }

    Region& createRegion(uint32 region_id)
    {
      ntk_assert(!ntk::has_key(m_regions, region_id), "Region already exists.");
      m_regions[region_id] = Region(region_id);
      return m_regions[region_id];
    }
    
    void removeRegion(const Region& r) { removeRegion(r.id()); }
    void removeRegion(uint32 region_id)
    {
      ntk_assert(ntk::has_key(m_regions, region_id), "Region does not exists.");
      m_regions.erase(region_id);
    }

    void updateRegion(const Region& r)
    {
      ntk_assert(ntk::has_key(m_regions, r.id()), "Region does not exists.");
      m_regions[r.id()] = r;
    }

    Region& getRegionRef(uint32 region_id)
    {
      ntk_assert(ntk::has_key(m_regions, region_id), "Region does not exists.");
      return m_regions[region_id];
    }
    
    void mergeR2IntoR1(uint32 r1_id, uint32 r2_id);
    
  protected:
    virtual void initialize();
    virtual void finalize() { finalizeImpl(); }
    
    virtual void segmentImpl() = 0;
    virtual void initializeImpl() {};
    virtual void finalizeImpl() {};
    void createRegionLinks();
    void createRegionLinksAndLabelImage();
    void importFromLabelImage(const vil_image_view<int32>& label_image);

  protected:
    vcl_map<uint32,Region> m_regions;
    ntk::vil_byte_image m_image_to_segment;
    vil_image_view<int32> m_label_image;
    vil_image_view<uint8> m_current_segmentation;
    int m_refresh_rate;
    CriticalSection m_lock;
};

/**
 * Functor which segments a given image into a label image, such that
 * different regions have different labels.
 */
class SegmentationFunctor
{
  public:
    virtual ~SegmentationFunctor() {}
    virtual String name() const = 0;
    virtual vil_image_view<int32> operator()(const vil_image_view<uint8>& im) = 0;
};

/**
 * Segment an image using an external functor.
 */
class ExternalSegmentor : public Segmentor
{
  public:
    /** 
     * Program will be run with two parameters : input file and output file :
   *   program input.pgm output.pgm
     */
    ExternalSegmentor(SegmentationFunctor* functor)
      : m_functor(functor)
    {
      ntk_assert(functor != 0, "A functor must be provided.");
    }
    
    virtual ~ExternalSegmentor() { delete m_functor; }

    virtual String name() const 
    { return String("External") + m_functor->name(); }
    
  protected:
    virtual void segmentImpl() 
    { 
      vil_image_view<int32> labels = (*m_functor)(getImageToSegment());
      importFromLabelImage(labels);
    }
    
  private:
    SegmentationFunctor* m_functor;
};

/**
 * Segment an image by regrouping adjacent identical pixels.
 */
class IdenticalPixelsFunctor : public SegmentationFunctor
{
  public:
    virtual ~IdenticalPixelsFunctor() {}

    virtual String name() const { return String("IdenticalPixels"); }
    
  protected:
    virtual vil_image_view<int32> operator()(const ntk::vil_byte_image& im)
    { 
      vil_image_view<int32> output;
      vil_convert_cast(im, output);
      return output;
    }
};

#endif // ndef ACSEGMENTOR_segmentor_H
