//
// watershed.cc
//
// Author: Nicolas Burrus <nicolas.burrus@ensta.fr>, (C) 2007
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
//

#include "watershed.h"

#include <ntk/ntk_with_macros.h>
#include <vcl_queue.h>
#include <vcl_vector.h>
#include <vil/vil_border.h>
#include <vil/algo/vil_gauss_filter.h>

using namespace ntk;

namespace
{
  const int32 wshed = vcl_numeric_limits<uint8>::max();
  const int32 init = vcl_numeric_limits<uint8>::max() - 1;
  const int32 mask = vcl_numeric_limits<uint8>::max() - 2;
  const int32 inqueue = vcl_numeric_limits<uint8>::max() - 3;
  const int32 maxlevel = inqueue - 1;

  struct watershed_con_point_handler
  {
    template<class Fifo, class II, class IO>
    static inline void
    process(Fifo& fifo, const II& im_i, IO& im_o)
    {
      typedef typename IO::pixel_type destvalue_type;
      (void) init; (void) wshed; /* unused */

      vil_point p = fifo.front();
      fifo.pop();

      if (im_o(p.x,p.y) == inqueue)
      {
        // Find the labeled neighbor with the lower level
        for (size_t idp = 0; idp < vil_neighb_c8().size(); ++idp)
        {
          vil_point np = p + vil_neighb_c8()[idp];
          if (!im_o.in_range(np.x, np.y)) continue;
          if (im_o(np.x, np.y) > maxlevel) continue;

          vil_point l = np;
          for (size_t jdp = idp; jdp < vil_neighb_c8().size(); ++jdp)
          {
            vil_point nnp = p + vil_neighb_c8()[jdp];
            if (!im_o.in_range(nnp.x, nnp.y)) continue;
            if (im_o(nnp.x, nnp.y) <= maxlevel
                && im_i(nnp.x, nnp.y) < im_i(l.x,l.y))
              l = nnp;
          }

          im_o(p.x,p.y) = im_o(l.x,l.y);
          break;
        }
      }

      /* Enqueue all masked neighbors */
      foreach_idx(idp, vil_neighb_c8())
      {
        vil_point np = p + vil_neighb_c8()[idp];
        if (!im_o.in_range(np.x, np.y)) continue;
        if (im_o(np.x, np.y) != mask) continue;
        im_o(np.x, np.y) = inqueue;
        fifo.push(np);
      }
    }
  };

  /*!
   ** \brief Check  if the second element  of p1 is  lower than the
   ** second one of p2.
   */
  template<class PointType, class ValueType> inline
  bool
  watershed_seg_sort(const vcl_pair<PointType, ValueType>& p1,
                     const vcl_pair<PointType, ValueType>& p2)
  {
    return p1.second < p2.second;
  }

  vil_image_view<int32>
  soille_watershed(const vil_byte_image& im_i)
  {
    // Initializations
    vil_image_view<int32> im_o(im_i.ni(), im_i.nj());
    im_o.fill(init);
    vcl_queue < vil_point > fifo;
    int32 current_label = 0;

    // Sort the pixels of im_i in the increasing order of their grey
    // values.
    vcl_vector < vcl_pair < vil_point, uint8 > > histo(im_i.ni()*im_i.nj());

    {
      unsigned int i = 0;
      for_all_xy(im_i) histo[i++] = vcl_make_pair(vil_point(x,y), im_i(x,y));

      // FIXME: We don't use a distrib_sort for now because that would
      // not work with float values.  We should have a specialized
      // sorting proc to handle this in the future.
      vcl_sort(histo.begin(), histo.end(),
               watershed_seg_sort<vil_point, uint8>);
    }

    for (unsigned int i = 0; i < histo.size();)
    // geodesic SKIZ of level h-1 inside level h
    {
      unsigned int level_start = i; // First index for current level.
      uint8 h = histo[i].second;
      vil_point p;

      while (i < histo.size())
      {
        p = histo[i].first;
        if (im_i(p.x,p.y) != h)
          break;

        im_o(p.x,p.y) = mask;

        bool flag_exist = false;
        foreach_idx(idp, vil_neighb_c8())
        {
          vil_point np = p + vil_neighb_c8()[idp];
          if (!im_o.in_range(np.x, np.y)) continue;
          if (im_o(np.x, np.y) == wshed || im_o(np.x, np.y) <= maxlevel)
            flag_exist = true;
        }
        if (flag_exist == true)
        {
          im_o(p.x,p.y) = inqueue;
          fifo.push(p);
        }

        ++i;
      }

      // ================================================================

      while (!fifo.empty())
        watershed_con_point_handler::process(fifo, im_i, im_o);

      // ================================================================

      /* Check for new minima.  */

      unsigned int j = level_start;
      while (j < histo.size())
      {
        p = histo[j].first;
        if (im_i(p.x,p.y) != h)
          break;

        if (im_o(p.x,p.y) == mask)
        {
          current_label += 1;
          /* If we run out of labels, restart from the initial
                          value.  FIXME: We should have a mean to tell
                          the caller that such 'wrapping' occured.  */
          if (current_label > maxlevel)
            current_label = 0;

          fifo.push(p);
          im_o(p.x,p.y) = current_label;
          while (!fifo.empty())
          {
            vil_point p_prime = fifo.front();
            fifo.pop();

            foreach_idx(idp, vil_neighb_c8())
            {
              vil_point np = p_prime + vil_neighb_c8()[idp];
              if (!im_o.in_range(np.x, np.y)) continue;
              if (im_o(np.x, np.y) != mask) continue;
              fifo.push(np);
              im_o(np.x, np.y) = current_label;
            }
          }
        }
        ++j;
      }
    }
    return im_o;
  }

} // namespace anonymous

vil_image_view<int32> watershed_segment(const ntk::vil_byte_image& im,
                                        double sigma)
{
  vil_byte_image smooth_im;
  vil_gauss_filter_5tap(im, smooth_im, vil_gauss_filter_5tap_params(sigma));

  vil_byte_image grad_magn;
  vil_sobel_intensity(smooth_im, grad_magn, vil_border_create_geodesic(im));

  vil_image_view<int32> labels = soille_watershed(grad_magn);
  return labels;
}
